#!/usr/bin/mzscheme -r
;
;  PRCSweb -- An HTML front-end for PRCS
; Released under the GNU GPL... see related README file
; Copyright (c) 2002 MJ Ray, markj at luminas.co.uk

(require
  (lib "13.ss" "srfi")
  (lib "pregexp.ss")
  (lib "process.ss"))

(putenv "PRCS_REPOSITORY" "/home/markj/PRCS/")
(putenv "PRCS_PLAIN_FORMAT" "1")

(define (shell-escape string)
  (pregexp-replace* "[\\*\\?;'&|\\(\\)\\<\\> 	]" string "\\\\\\0"))

(define (get-output command) ; -> list-line-strings
  (apply
    (lambda (out in pid err status)
      (let gather ((outs '()))
        (if (equal? eof (peek-char out))
            (reverse outs)
            (gather (cons (read-line out) outs)))))
    (process command)))

(define html-header #f)
(define html-footer #f)

(let ((headerfile ""))
  (set! html-header
    (lambda (title)
      (display "Content-type: text/html\n\n")
      (set! headerfile (pregexp-match "^([[:^word:][:word:]]*)(<\\?prcsweb \\?>)([[:^word:][:word:]]*)" (call-with-input-file "prcsweb.html" (lambda (p) (read-string 65535 p)))))
      (display (format (cadr headerfile) title))))
  (set! html-footer (lambda () (display (cadddr headerfile)))))

(define (do-html-diff proj maj1 min1 maj2 min2 file)
  ; checkout file1 and file2
  (get-output (format "prcs execute -r~S.~S '--match=^~A$' -f ~S -- cat '{file}' >/tmp/prcsweb/version_~S.~S" maj1 min1 file proj maj1 min1))
  (get-output (format "prcs execute -r~S.~S '--match=^~A$' -f ~S -- cat '{file}' >/tmp/prcsweb/version_~S.~S" maj2 min2 file proj maj2 min2))
  (display (apply string-append (get-output (format "python diff2html /tmp/prcsweb/version_~S.~S /tmp/prcsweb/version_~S.~S" maj1 min1 maj2 min2)))))

(define (project-list)
  (let ((projects (get-output "prcs admin pinfo -f -q")))
    (display "<h1>Projects</h1><ul>")
    (for-each (lambda (p)
                (display (format "<li><a href='~A/'>~A</a>\n" p p)))
      projects)
    (display "</ul>")))

(define (branch-list project)
  (let ((branches (get-output (format "prcs info -r*.@ -f ~S 2>&1" (shell-escape project)))))
    (display "<h1>Branches (latest entries)</h1><ul>")
    (for-each (lambda (p)
                (let ((ver (car (cdr (string-tokenize p)))))
                  (display (format "<li><a href='~A/'>~A</a>\n" (substring ver 0 (string-index-right ver #\.)) p))))
      branches)
    (display (format "</ul><a href='../../maps/~A'>Postscript map</a>" project))))

(define (version-list project branch)
  (let ((branches (get-output (format "prcs info -r~S -f ~S 2>&1" (shell-escape branch) (shell-escape project)))))
    (display "<h1>Versions</h1><ul>")
    (for-each (lambda (p)
                (let ((ver (car (cdr (string-tokenize p)))))
                  (display (format "<li><a href='~A/'>~A</a>\n" (substring ver (+ (string-index-right ver #\.) 1) (string-length ver)) p))))
      branches)
    (display "</ul>")))

(define (file-list diff project branch version)
  (display (format "<h1>Version ~A.~A</h1><pre>~A</pre>" branch version (string-join (get-output (format "prcs changes -l -r ~S.~S -f ~S" (shell-escape branch) (shell-escape version) (shell-escape project))) "\n")))
  (let ((rellink (if diff "" (format "~A/files/~A/~A/~A/" (getenv "SCRIPT_NAME") project branch version)))
        (files (get-output (format "prcs execute -r~S.~S -f ~S -- echo '{}'" (shell-escape branch) (shell-escape version) (shell-escape project)))))
    (display (format "<p><a href='~A/diffs/~A/~A/~A/'>Diff this with another version</a></p><h2>Files</h2><ul>" (getenv "SCRIPT_NAME") project branch version))
    (for-each (lambda (p)
                (display (format "<li><a href='~A~A'>~A</a>\n" rellink p p)))
      files)
    (display "</ul>")))

(define (view-file project branch version . path)
  (display (format "Content/type: text/plain\n\n~A" (string-join (get-output (format "prcs execute -r~S.~S -f '--match=^~A$' ~S -- cat '{file}' 2>&1" (shell-escape branch) (shell-escape version) (shell-escape (string-join path "/")) (shell-escape project))) "\n"))))

(define (project-map project)
  (display "Content-type: application/postscript\n\n")
  (system (format "rm /tmp/prcsweb/prcsweb.ps ; prcs info -l -f ~S | prcs_tree_info | xvcg - -psoutput /tmp/prcsweb/prcsweb.ps >/dev/null ; cat /tmp/prcsweb/prcsweb.ps" (shell-escape project))))

(define (project-view . path)
  (let ((patlen (length path)))
  (html-header path)
  (cond
        ((null? path) (project-list))
        ((= patlen 1) (apply branch-list path))
        ((= patlen 2) (apply version-list path))
        ((= patlen 3) (apply file-list (cons #f path)))
        (#t "Unhandled request."))
  (html-footer)))

(define (project-diff . path)
  (let ((proj (car path))
        (maj (cadr path))
        (min (caddr path))
        (path (cdddr path)))
    (cond
          ((null? path)
             (html-header (list proj maj min path))
             (display "<h1>Pick branch to diff against</h1>")
             (branch-list proj)
             (html-footer))
          ((null? (cdr path))
             (html-header (list proj maj min path))
             (display "<h1>Pick version to diff against</h1>")
            (apply version-list (cons proj path))
            (html-footer))
          ((null? (cddr path))
             (html-header (list proj maj min path))
             (display (format "<h1>Diffing Version ~A.~A Against Version ~A.~A</h1><p><a href='../../../../~A/~A/~A/~A/'>Reverse diff versions</a>" maj min (car path) (cadr path) (car path) (cadr path) maj min))
             (display "<table><tr valign='top'><td>")
             (apply file-list (cons #t (map shell-escape (list proj maj min))))
             (display "</td><td>")
             (apply file-list (cons #t (map shell-escape (cons proj path))))
             (display "</td></tr></table>")
             (html-footer))
          (#t
              (display "Content-type: text/html\n\n")
              (apply do-html-diff (map shell-escape (list proj maj min (car path) (cadr path) (string-join (cddr path) "/"))))))))

(let* ((path-info (getenv "PATH_INFO"))
       (path (string-tokenize path-info (char-set-complement (list->char-set '(#\/))))))
  (cond
    ((pregexp-match "\\.\\./" path-info) (display "Location: ../\n\n"))
    ((null? path) (display "Location: ./projects/\n\n"))
    ((equal? (car path) "projects") (apply project-view (cdr path)))
    ((equal? (car path) "files") (apply view-file (cdr path)))
    ((equal? (car path) "diffs") (apply project-diff (cdr path)))
    ((equal? (car path) "maps") (apply project-map (cdr path)))
    (#t (display "Content-type: text/plain\n\nUnhandled request."))))
